"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExprContext = void 0;
var constants_1 = require("../constants");
var boolean_value_1 = require("./values/boolean-value");
var node_set_value_1 = require("./values/node-set-value");
var number_value_1 = require("./values/number-value");
var string_value_1 = require("./values/string-value");
var tokens_1 = require("./tokens");
/**
 * XPath expression evaluation context. An XPath context consists of a
 * DOM node, a list of DOM nodes that contains this node, a number
 * that represents the position of the single node in the list, and a
 * current set of variable bindings. (See XPath spec.)
 *
 *   setVariable(name, expr) -- binds given XPath expression to the
 *   name.
 *
 *   getVariable(name) -- what the name says.
 *
 *   setNode(position) -- sets the context to the node at the given
 *   position. Needed to implement scoping rules for variables in
 *   XPath. (A variable is visible to all subsequent siblings, not
 *   only to its children.)
 *
 *   set/isCaseInsensitive -- specifies whether node name tests should
 *   be case sensitive.  If you're executing xpaths against a regular
 *   HTML DOM, you probably don't want case-sensitivity, because
 *   browsers tend to disagree about whether elements & attributes
 *   should be upper/lower case.  If you're running xpaths in an
 *   XSLT instance, you probably DO want case sensitivity, as per the
 *   XSL spec.
 *
 *   set/isReturnOnFirstMatch -- whether XPath evaluation should quit as soon
 *   as a result is found. This is an optimization that might make sense if you
 *   only care about the first result.
 *
 *   set/isIgnoreNonElementNodesForNTA -- whether to ignore non-element nodes
 *   when evaluating the "node()" any node test. While technically this is
 *   contrary to the XPath spec, practically it can enhance performance
 *   significantly, and makes sense if you a) use "node()" when you mean "*",
 *   and b) use "//" when you mean "/descendant::* /".
 */
var ExprContext = /** @class */ (function () {
    /**
     * Constructor -- gets the node, its position, the node set it
     * belongs to, and a parent context as arguments. The parent context
     * is used to implement scoping rules for variables: if a variable
     * is not found in the current context, it is looked for in the
     * parent context, recursively. Except for node, all arguments have
     * default values: default position is 0, default node set is the
     * set that contains only the node, and the default parent is null.
     *
     * Notice that position starts at 0 at the outside interface;
     * inside XPath expressions this shows up as position()=1.
     * @param nodeList TODO
     * @param outputNodeList TODO
     * @param opt_position TODO
     * @param opt_outputPosition TODO
     * @param opt_parent TODO
     * @param opt_caseInsensitive TODO
     * @param opt_ignoreAttributesWithoutValue TODO
     * @param opt_returnOnFirstMatch TODO
     * @param opt_ignoreNonElementNodesForNTA TODO
     */
    function ExprContext(nodeList, outputNodeList, xsltVersion, opt_position, opt_outputPosition, opt_outputDepth, opt_decimalFormatSettings, opt_variables, opt_knownNamespaces, opt_parent, opt_caseInsensitive, opt_ignoreAttributesWithoutValue, opt_returnOnFirstMatch, opt_ignoreNonElementNodesForNTA) {
        if (xsltVersion === void 0) { xsltVersion = '1.0'; }
        this.nodeList = nodeList;
        this.outputNodeList = outputNodeList;
        this.xsltVersion = xsltVersion;
        this.position = opt_position || 0;
        this.outputPosition = opt_outputPosition || 0;
        this.variables = opt_variables || {};
        this.keys = (opt_parent === null || opt_parent === void 0 ? void 0 : opt_parent.keys) || {};
        this.knownNamespaces = opt_knownNamespaces || {};
        this.parent = opt_parent || null;
        this.caseInsensitive = opt_caseInsensitive || false;
        this.ignoreAttributesWithoutValue = opt_ignoreAttributesWithoutValue || false;
        this.returnOnFirstMatch = opt_returnOnFirstMatch || false;
        this.ignoreNonElementNodesForNTA = opt_ignoreNonElementNodesForNTA || false;
        this.inApplyTemplates = false;
        this.baseTemplateMatched = false;
        this.outputDepth = opt_outputDepth || 0;
        this.decimalFormatSettings = opt_decimalFormatSettings || {
            decimalSeparator: '.',
            groupingSeparator: ',',
            infinity: 'Infinity',
            minusSign: '-',
            naN: 'NaN',
            percent: '%',
            perMille: '‰',
            zeroDigit: '0',
            digit: '#',
            patternSeparator: ';'
        };
        if (opt_parent) {
            this.root = opt_parent.root;
        }
        else if (this.nodeList[this.position].nodeType == constants_1.DOM_DOCUMENT_NODE) {
            // NOTE(mesch): DOM Spec stipulates that the ownerDocument of a
            // document is null. Our root, however is the document that we are
            // processing, so the initial context is created from its document
            // node, which case we must handle here explicitly.
            this.root = this.nodeList[this.position];
        }
        else {
            this.root = this.nodeList[this.position].ownerDocument;
        }
    }
    /**
     * clone() -- creates a new context with the current context as
     * parent. If passed as argument to clone(), the new context has a
     * different node, position, or node set. What is not passed is
     * inherited from the cloned context.
     * @param opt_nodeList TODO
     * @param opt_outputNodeList TODO
     * @param opt_position TODO
     * @param opt_outputPosition TODO
     * @returns TODO
     */
    ExprContext.prototype.clone = function (opt_nodeList, opt_outputNodeList, opt_position, opt_outputPosition) {
        return new ExprContext(opt_nodeList || this.nodeList, opt_outputNodeList || this.outputNodeList, this.xsltVersion, typeof opt_position !== 'undefined' ? opt_position : this.position, typeof opt_outputPosition !== 'undefined' ? opt_outputPosition : this.outputPosition, this.outputDepth, this.decimalFormatSettings, this.variables, this.knownNamespaces, this, this.caseInsensitive, this.ignoreAttributesWithoutValue, this.returnOnFirstMatch, this.ignoreNonElementNodesForNTA);
    };
    ExprContext.prototype.cloneByOutput = function (opt_outputNodeList, opt_outputPosition, opt_outputDepth) {
        return new ExprContext(this.nodeList, opt_outputNodeList || this.outputNodeList, this.xsltVersion, this.position, typeof opt_outputPosition !== 'undefined' ? opt_outputPosition : this.outputPosition, typeof opt_outputDepth !== 'undefined' ? opt_outputDepth : this.outputDepth, this.decimalFormatSettings, this.variables, this.knownNamespaces, this, this.caseInsensitive, this.ignoreAttributesWithoutValue, this.returnOnFirstMatch, this.ignoreNonElementNodesForNTA);
    };
    ExprContext.prototype.setVariable = function (name, value) {
        if (value instanceof string_value_1.StringValue ||
            value instanceof boolean_value_1.BooleanValue ||
            value instanceof number_value_1.NumberValue ||
            value instanceof node_set_value_1.NodeSetValue) {
            this.variables[name] = value;
            return;
        }
        if ('true' === value) {
            this.variables[name] = new boolean_value_1.BooleanValue(true);
        }
        else if ('false' === value) {
            this.variables[name] = new boolean_value_1.BooleanValue(false);
        }
        else if (tokens_1.TOK_NUMBER.re.test(String(value))) {
            this.variables[name] = new number_value_1.NumberValue(value);
        }
        else {
            // DGF What if it's null?
            this.variables[name] = new string_value_1.StringValue(value);
        }
    };
    ExprContext.prototype.getVariable = function (name) {
        if (typeof this.variables[name] != 'undefined') {
            return this.variables[name];
        }
        if (this.parent) {
            return this.parent.getVariable(name);
        }
        return null;
    };
    ExprContext.prototype.setNode = function (position) {
        this.position = position;
    };
    ExprContext.prototype.contextSize = function () {
        return this.nodeList.length;
    };
    ExprContext.prototype.isCaseInsensitive = function () {
        return this.caseInsensitive;
    };
    ExprContext.prototype.setCaseInsensitive = function (caseInsensitive) {
        return (this.caseInsensitive = caseInsensitive);
    };
    ExprContext.prototype.isIgnoreAttributesWithoutValue = function () {
        return this.ignoreAttributesWithoutValue;
    };
    ExprContext.prototype.setIgnoreAttributesWithoutValue = function (ignore) {
        return (this.ignoreAttributesWithoutValue = ignore);
    };
    ExprContext.prototype.isReturnOnFirstMatch = function () {
        return this.returnOnFirstMatch;
    };
    ExprContext.prototype.setReturnOnFirstMatch = function (returnOnFirstMatch) {
        return (this.returnOnFirstMatch = returnOnFirstMatch);
    };
    ExprContext.prototype.isIgnoreNonElementNodesForNTA = function () {
        return this.ignoreNonElementNodesForNTA;
    };
    ExprContext.prototype.setIgnoreNonElementNodesForNTA = function (ignoreNonElementNodesForNTA) {
        return (this.ignoreNonElementNodesForNTA = ignoreNonElementNodesForNTA);
    };
    return ExprContext;
}());
exports.ExprContext = ExprContext;
//# sourceMappingURL=expr-context.js.map